var path = require('path');
var moment = require('moment');
var sequelize = require('sequelize');
var IOService = require("../components/services/io");
var env = process.env.NODE_ENV || "prod";
var configJson = require(path.join(__dirname, '..', '/components/config/config.json'))[env];
var AsnHelper = require("../components/helpers/AsnHelper");

var fs = require('fs');
var _ = require("lodash");

// FRES-170 - global variable -flag- to store the need to rebuild the routing table
global.buildRouting = false;
// 1 hour -> 3600000 ms
// 15 minutes -> 900000 ms
// 30 minutes -> 1800000 ms
// 2 minutes -> 120000 ms
var buildRoutingTableMillisec = 1800000;  // Check every 30 minutes
var checkingUTCTime = 7;                  // Checking for 7 AM UTC - equivalent to 2 AM ET

module.exports = function(app, passport, models) {
    var winston = app.get('winston');
    var logger = winston.loggers.get('main');

    app.get("/api/healthcheck", function (req, res) {
        res.status(200).send("OK");
    });

    // API entry to report back the server version.
    // Used by the app to determine if the server support specific functions
    app.get("/api/serverVersion", function (req, res) {
        res.send("1.0.113");        // Retrieve version here ? TBD
    });

    //APIs for Mobile Application
    app.post('/api/saveCountData', function (req, res) {
        bulkCreateData(req.body, 'count')
            .done(function (data) {
                res.send(data);
            });
    });

    app.post('/api/saveInventoryData', function (req, res) {
        bulkCreateData(req.body, 'inventory')
            .done(function (data) {
                res.send(data);
            });
    });


    //MASTER ITEMs

    // -------------------------------------------------------------------  SAP Keys management -- START
    // TFG - FRES-116 - API to retrieve SFTP keys for SAP
    // The keys are generated manually via ssh-keygen and saved in the directory /home/centos/sap_sftp_keys,
    // named sfpSap (private key) and sftSap.pub (public key).
    // See documentation for the associated passphrase and how to generate the keys

    app.get('/api/setSftpSapKeys', function (req, res) {
        logger.info('API setSftpSapKeys', 'Called');

        // At the date (Sept 2019) the API exists but is doing nothing.
        // For future use.

        res.send("No key generated yet - TBD");
    });

    app.get('/api/getSftpSapPuk', function (req, res) {
        logger.info('API getSftpSapPuk', 'Called');

        // At the date (Sept 2019) this API return the public key to be added to the SFTP server.
        // Note !! The public key is returned in clear ! No JSON formatting.
        // Use curl to retrieve it - the devices are NOT asking for this !

        res.sendFile('sap_sftp_keys/sftpSap.pub', {root: path.dirname(__dirname)});
    });

    app.get('/api/getSftpSapPrk', function (req, res) {
        logger.info('API getSftpSapPrk', 'Called');

        // At the date (Sept 2019) this API return the stored private key file MANUALLY edited
        // The sendFile call send whatever the file is, no JSON formatting added, thus the file to be sent
        // must be edited.
        // Currently the file is manually formatted JSON - see confluence for details
        res.sendFile('sap_sftp_keys/sftpSap.prep', {root: path.dirname(__dirname)});
    });
    // -------------------------------------------------------------------  SAP Keys management ---- END

    // -------------------------------------------------------------------  Tables management -- START
    /**
     * Set global flag to notify the need to rebuild the routing table
     */
    app.post('/api/setFlagNeedRoutingRebuilding', function (req, res) {
        logger.info('/api/setFlagNeedRoutingRebuilding', 'Called');
        global.buildRouting = true;

        res.send("Ok");
    });

// FRES-170 - settign up call to rebuild the routing table every day at 1 am
    setInterval(function() {
        var objDate = new Date();
        var hours = objDate.getHours();
        if(hours === checkingUTCTime ){
            logger.info("--> Rebuild routing table if flag set !");

            if (global.buildRouting === true) {
                logger.info("--> Rebuilding routing table");
                // Global flag to rebuild reset in the rebuild function if all OK

                rebuildRoutingTable();
            }

            // Note - this can be called twice a day around 2 AM - no check on minutes - only hour
            // At May 29 2020 don't activate it - need discussion
            //logger.info("--> Purge Activity log table from 182 days older records");
            //purgeOldActivityLogEntries();

        } else{
            // Shorter log to don't add too much to the log
            logger.info(". RR check");
        }
    }, buildRoutingTableMillisec);

    /**
     * Returns all VendorDestination in the table
     */
    app.get('/api/v1/getVendorsRouteList', function (req, res) {
        logger.info('API /v1/getVendorsRouterList', 'API for app < 1.6.15');

        var queryInsert = "insert into `tbl_vendor_route_product` (sap_material_id, vendor_material, vendor_id, vendor_route) \n " +
            "select sap_material_id, vendor_material, vendor_id, (select vendor_route from tbl_vendor_route where vendor_id = tbl_master_item.vendor_id) \n" +
            "from `tbl_master_item` where discontinued IS NULL or discontinued = ''";

        // First erase the current table
        logger.info('API /v1/getVendorsRouteList', 'Erase vendor product table');
        var query = "delete from `tbl_vendor_route_product`";
        models.sequelize.query(query).spread(function (results, metadata) {
            logger.info('API /v1/getVendorsRouteList', 'Reset index');
            query = "alter table `tbl_vendor_route_product` auto_increment=1";
            models.sequelize.query(query).spread(function (results, metadata) {
                logger.info('API /v1/getVendorsRouteList', 'Rebuild vendor product table');
                // Rebuild the table
                models.sequelize.query(queryInsert).spread(function (results, metadata) {
                    logger.info('API /v1/getVendorsRouteList', 'Send result to terminal');
                    // Return to the caller the table
                    models.VendorRouteProduct.findAll({order: '"vendorMaterial" ASC'}).then(function (vendors) {
                        res.send(vendors);
                    });
                }).catch(function (insErr) {
                    logger.info('API /v1/getVendorsRouteList', '- ERROR on insert - ' + insErr);
                    res.sendStatus(500);
                });
            }).catch(function (insErr) {
                logger.info('API /v1/getVendorsRouteList', '- ERROR on reset index - ' + insErr);
                res.sendStatus(500);
            });
        }).catch(function (insErr) {
            logger.info('API /v1/getVendorsRouteList', '- ERROR on delete - ' + insErr);
            res.sendStatus(500);
        });
    });

    /**
     * FRES-149
     * Returns all VendorDestination in the table for a specific clinic
     *
     *  CAUTION !!!!  This code is conflict prone in case of multiple log in from different terminals.
     *  Once generated the table, the table is sent to the terminal and it can takes time OUTSIDE this thread.
     *  Another thread can call it if anoother login happens erasing the table and rendering the one still sending invalid.
     */
    app.get('/api/v2/getVendorsRouteList/:clinic', function (req, res) {
        var queryInsert = "";
        if (!req.params.clinic) {
            logger.info('API /v2/getVendorsRouteList', 'API for app >= 1.6.15 - Called with no clinic ID ');
            res.status(400).send("Missing Clinic ID");
        } else {
            logger.info('API /v2/getVendorsRouteList', 'API for app >= 1.6.15 - Called for clinic ' + req.params.clinic);

            locClinicId = req.params.clinic;

            queryInsert = "insert into `tbl_vendor_route_product` (sap_material_id, vendor_material, vendor_id, vendor_route) \n " +
                "select sap_material_id, vendor_material, vendor_id, \n" +
                "if (exists (select vendor_route from tbl_vendor_route_clinic where (vendor_id = tbl_master_item.vendor_id and clinic_id = '" + locClinicId + "')), \n" +
                "(select vendor_route from tbl_vendor_route_clinic where (vendor_id = tbl_master_item.vendor_id and clinic_id = '" + locClinicId + "')), \n" +
                "(select vendor_route from tbl_vendor_route where vendor_id = tbl_master_item.vendor_id)) \n" +
                "from `tbl_master_item` where discontinued IS NULL or discontinued = ''";

            // First erase the current table
            logger.info('API /v2/getVendorsRouteList', 'Erase vendor product table');
            var query = "delete from `tbl_vendor_route_product`";
            models.sequelize.query(query).spread(function (results, metadata) {
                logger.info('API /v2/getVendorsRouteList', 'Reset index');
                query = "alter table `tbl_vendor_route_product` auto_increment=1";
                models.sequelize.query(query).spread(function (results, metadata) {
                    logger.info('API /v2/getVendorsRouteList', 'Rebuild vendor product table');
                    // Rebuild the table
                    models.sequelize.query(queryInsert).spread(function (results, metadata) {
                        logger.info('API /v2/getVendorsRouteList', 'Send result to terminal');
                        // Return to the caller the table
                        models.VendorRouteProduct.findAll({order: '"vendorMaterial" ASC'}).then(function (vendors) {
                            res.send(vendors);
                        });
                    }).catch(function (insErr) {
                        logger.info('API /v2/getVendorsRouteList', '- ERROR on insert - ' + insErr);
                        res.sendStatus(500);
                    });
                }).catch(function (insErr) {
                    logger.info('API /v2/getVendorsRouteList', '- ERROR on reset index - ' + insErr);
                    res.sendStatus(500);
                });
            }).catch(function (insErr) {
                logger.info('API /v2/getVendorsRouteList', '- ERROR on delete - ' + insErr);
                res.sendStatus(500);
            });
        }
    });

    /**
     * FRES-169
     * Returns all VendorDestination in the table for a specific clinic
     * This code uses a STATIC database containing the routing for the clinics. NOt all the clinics are in the routing table,
     * thus if a clinic is not present, a global routing is sent instead (uses the client_id .0000BM)
     * The .0000BM routing table is the global one
     */
    app.get('/api/v3/getVendorsRouteList/:clinic', function (req, res) {
        if (!req.params.clinic) {
            logger.info('API /v3/getVendorsRouteList', 'API for app >= 1.7.0 - Called with no clinic ID ');
            res.status(400).send("Missing Clinic ID");
        } else {
            logger.info('API /v3/getVendorsRouteList', 'API for app >= 1.7.0 - Called for clinic ' + req.params.clinic);

            var locClinicId = req.params.clinic;
            var whereArg = [];

            whereArg.push({clinic_id: locClinicId});    // Assign clinic_id for the retrieval condition
            models.VendorRouteApp.findAll({
                where: whereArg
            }).then(function (routes) {
                if (routes === null || routes.length < 2) {
                    logger.debug('API /v3/getVendorsRouteList', 'Retrieve entry failed - try for .0000BM');
                    whereArg.pop();
                    whereArg.push({clinic_id: ".0000BM"});

                    models.VendorRouteApp.findAll({
                        where: whereArg
                    }).then(function (routesAlt) {
                        if (routesAlt == null) {
                            // BAD BAD BAD !
                            logger.error('API /v3/getVendorsRouteList', 'Retrieve entry failed for .0000BM !!');
                            res.sendStatus(500);
                        } else {
                            logger.info('API /v3/getVendorsRouteList', 'Sending entry : ' + JSON.stringify(routesAlt, null, 4));
                            res.send(routesAlt);
                        }
                    }).catch(function (err) {
                        logger.error(err);
                        res.sendStatus(500);
                    });
                } else {
                    logger.info('API /v3/getVendorsRouteList', 'Sending entry : ' + JSON.stringify(routes, null, 4));
                    res.send(routes);
                }
            }).catch(function (err) {
                logger.error(err);
                res.sendStatus(500);
            });
        }
    });

    app.post('/api/saveRouteVendor', function (req, res) {
        if (req.body.id && req.body.id > 0) {
            models.VendorRoute.update(req.body, {
                where: {
                    id: req.body.id
                }
            }).then(function (count) {
                res.send(req.body);
            }).catch(function (err) {
                logger.error(err);
                res.sendStatus(500);
            });
        } else {
            if (req.body.vendorId) {
                models.VendorRoute.findOrCreate({
                    where: {
                        vendorId: req.body.vendorId
                    }, defaults: {
                        vendorPathRoute: req.body.vendorPathRoute,
                    }
                }).spread(function (data, created) {
                    if (!created) {
                        data.vendorPathRoute = req.body.vendorPathRoute;
                        data.save().then(function () {
                            res.send(data);
                        }).catch(function (error) {
                            logger.error(error);
                            res.sendStatus(500);
                        });
                    } else {
                        res.send(data);
                    }
                }).catch(function (err) {
                    logger.error(err);
                    res.sendStatus(500);
                });
            } else {
                res.sendStatus(500);
            }
        }
    });

    app.post('/api/deleteRouteVendor', function (req, res) {
        if (req.body.vendorId) {
            //TODO: make sure no Clinic or Item has this vendor assigned
            models.VendorRoute.destroy({
                where: req.body
            })
                .then(function () {
                    logger.info(process.pid + ":Vendor item deleted by: " + req.user.username + " record: " + JSON.stringify(req.body));
                    res.send(req.body);
                });
        } else {
            res.sendStatus(500);
        }
    });

    app.post('/api/saveRouteClinicVendor', function (req, res) {
        if (req.body.id && req.body.id > 0) {
            models.VendorRouteClinic.update(req.body, {
                where: {
                    id: req.body.id
                }
            }).then(function (count) {
                res.send(req.body);
            }).catch(function (err) {
                logger.error(err);
                res.sendStatus(500);
            });
        } else {
            if (req.body.vendorId && req.body.clinicId) {
                models.VendorRouteClinic.findOrCreate({
                    where: {
                        vendorId: req.body.vendorId,
                        clinicId: req.body.clinicId
                    }, defaults: {
                        vendorPathRoute: req.body.vendorPathRoute,
                    }
                }).spread(function (data, created) {
                    if (!created) {
                        data.vendorPathRoute = req.body.vendorPathRoute;
                        data.save().then(function () {
                            res.send(data);
                        }).catch(function (error) {
                            logger.error(error);
                            res.sendStatus(500);
                        });
                    } else {
                        logger.info(process.pid + ":Vendor route clinic item added by: " + req.user.username + " record: " + JSON.stringify(req.body));
                        res.send(data);
                    }
                }).catch(function (err) {
                    logger.error(err);
                    res.sendStatus(500);
                });
            } else {
                res.sendStatus(500);
            }
        }
    });

    app.post('/api/deleteRouteClinicVendor', function (req, res) {
        if (req.body.vendorId && req.body.clinicId) {
            //TODO: make sure no Clinic or Item has this vendor assigned
            models.VendorRouteClinic.destroy({
                where: req.body
            })
                .then(function () {
                    logger.info(process.pid + ":Vendor route clinic item deleted by: " + req.user.username + " record: " + JSON.stringify(req.body));
                    res.send(req.body);
                });
        } else {
            res.sendStatus(500);
        }
    });

    app.post('/api/saveParVendor', function (req, res) {
        logger.debug("API saveParVendor ");     // TEMP
        if (req.body.id && req.body.id > 0) {
            models.VendorSapParValue.update(req.body, {
                where: {
                    id: req.body.id
                }
            }).then(function (count) {
                res.send(req.body);
            }).catch(function (err) {
                logger.error(err);
                res.sendStatus(500);
            });
        } else {
            if (req.body.vendorMaterial && req.body.vendorId && req.body.clinicId) {
                models.VendorSapParValue.findOrCreate({
                    where: {
                        vendorMaterial: req.body.vendorMaterial,
                        vendorId: req.body.vendorId,
                        clinicId: req.body.clinicId
                    }, defaults: {
                        minParValue: req.body.minParValue,
                        maxParValue: req.body.maxParValue,
                        qohValue: req.body.qohValue
                    }
                }).spread(function (data, created) {
                    logger.debug("   new record");     // TEMP

                    if (!created) {
                        data.vendorMaterial = req.body.vendorMaterial;
                        data.vendorId = req.body.vendorId;
                        data.clinicId = req.body.clinicId;
                        data.description = req.body.description;
                        data.minParValue = req.body.minParValue;
                        data.maxParValue = req.body.maxParValue;
                        data.qohValue = req.body.qohValue;

                        data.save().then(function () {
                            res.send(data);
                        }).catch(function (error) {
                            logger.error(error);
                            res.sendStatus(500);
                        });
                    } else {
                        res.send(data);
                    }
                }).catch(function (err) {
                    logger.error(err);
                    res.sendStatus(500);
                });
            } else {
                res.sendStatus(500);
            }
        }
    });

    app.post('/api/deleteParVendor', function (req, res) {
        if (req.body.vendorMaterial) {
            models.VendorSapParValue.destroy({
                where: req.body
            })
                .then(function () {
                    logger.info(process.pid + ":Vendor par value deleted by: " + req.user.username + " record: " + JSON.stringify(req.body));
                    res.send(req.body);
                });
        } else {
            res.sendStatus(500);
        }
    });

    /**
     * Verify that the new entry is a valid one.
     * - the vendor_id must exists in the tbl_list_table
     * - the vendor_material + vendor_id must exist in the tbl_master_item table
     * - the clinic_id must exists in the tbl_clinic_profile table
     */
    app.post('/api/verifyParVendor', function (req, res) {
        logger.info('API verifyParVendor', 'Called');

        var query = "TBD";

        models.sequelize.query(query).spread(function (results, metadata) {
            logger.info('API verifyParVendor', 'Checked ' + results);
        }).catch(function (insErr) {
            logger.info('API verifyParVendor', ' - ERROR adding missing rows in tbl_sap_item_par_value - ' + insErr);
            res.sendStatus(500);
            return;
        });

        res.sendStatus(200);
    });

    /**
     * Returns all VendorProducts in the table - the API is called from the app.
     */
    app.get('/api/v1/getVendorsProdList/:clinic', function (req, res) {
        if (!req.params.clinic) {
            res.status(400).send("Missing Clinic ID");
        } else {
            logger.info('API getVendorsProdList', 'Called for clinic ' + req.params.clinic);

            var clinicId = req.params.clinic;

            var where = [];
            if (clinicId != null) {
                where.push({clinic_id: clinicId});
                models.VendorSapParValue.findAll({
                    where: where
                }).then(function (profile) {
                    res.send(profile);
                });
            } else {
                res.sendStatus(500);
            }
        }
    });

    /**
     * Populate the tbl_sap_item_par_value table with the material description and vendor name
     */
    app.post('/api/v1/syncVendorParTable', function (req, res) {
        var query = "update `tbl_sap_item_par_value` \n" +
            "set vendor_display = (select vendor_display from `tbl_vendor_list` where tbl_sap_item_par_value.vendor_id = vendor_id), \n" +
            "material_description = (select distinct material_description from `tbl_master_item` \n" +
            "where (concat (tbl_sap_item_par_value.vendor_material, tbl_sap_item_par_value.vendor_id) = concat(vendor_material, vendor_id)) \n" +
            "and discontinued is null) ";

        models.sequelize.query(query).spread(function (results, metadata) {
            logger.info('API syncVendorParTable', 'Populated fields : ' + results);
        }).catch(function (insErr) {
            logger.info('API syncVendorParTable', ' - ERROR updating rows in tbl_sap_item_par_value - ' + insErr);
            res.sendStatus(500);
            return;
        });

        res.sendStatus(200);
    });

    /**
     * Syncronize the tbl_vendor_list table and the tbl_vendor_route
     * Note that the tbl_vendor_route MUST BE CREATED MANUALLY once (see queries)
     */
    app.post('/api/v1/syncVendorRouteTables', function (req, res) {
        logger.info('API syncVendorRouteTables', 'Called');

        // First step - add missing rows from tbl_vendor_list into tbl_vendor_route
        // It is by default assumed that the new vendor is routed to the IO - i.e. IO is the default
        var query = "insert into `tbl_vendor_route` \n" +
            "(vendor_id, vendor_display, vendor_route) select a.vendor_id, a.vendor_display, 'IO' \n" +
            "from `tbl_vendor_list` as a \n" +
            "where not exists (select vendor_id from tbl_vendor_route as b \n"+
            "where a.vendor_id = b.vendor_id)";
        models.sequelize.query(query).spread(function (results, metadata) {
            logger.info('API syncVendorRouteTables', 'Added missing rows in tbl_vendor_route when compared with tbl_vendor_list : ' + results);

            // Second step - remove missing rows from tbl_vendor_par_value NOT present in the tbl_master_item
            query = "delete from `tbl_vendor_route` \n" +
                "where not exists (select vendor_id from tbl_vendor_list as b \n" +
                "where tbl_vendor_route.vendor_id = b.vendor_id);";
            models.sequelize.query(query).spread(function (results, metadata) {
                logger.info('API syncVendorRouteTables', 'removed extra rows in tbl_vendor_route when compared with tbl_vendor_list : ' + results.status);

                // Third step - update table tbl_vendor_route_clinic filling up the vendor_display field
                query = "update `tbl_vendor_route_clinic` \n" +
                    "set  vendor_display = if (exists (select vendor_display from tbl_vendor_list where tbl_vendor_route_clinic.vendor_id = vendor_id), \n" +
                    " (select vendor_display from tbl_vendor_list where tbl_vendor_route_clinic.vendor_id = vendor_id), 'Invalid vendor')";
                models.sequelize.query(query).spread(function (results, metadata) {
                    logger.info('API syncVendorRouteTables', 'update vendor description on tbl_vendor_route_clinic : ' + results.status);
                    res.sendStatus(200);
                    return;
                }).catch(function (insErr) {
                    logger.info('API syncVendorRouteTables', ' - ERROR update vendor description on tbl_vendor_route_clinic - ' + insErr);
                    res.sendStatus(500);
                    return;
                });
            }).catch(function (insErr) {
                logger.info('API syncVendorRouteTables', ' - ERROR removing extra rows in tbl_vendor_route - ' + insErr);
                res.sendStatus(500);
                return;
            });
        }).catch(function (insErr) {
            logger.info('API syncVendorRouteTables', ' - ERROR adding missing rows in tbl_vendor_route - ' + insErr);
            res.sendStatus(500);
            return;
        });
    });

    /**
     * Rebuild the tbl_vendor_route_product table - the API is usually called from webconsole when the tbl_vendor_route
     * is changed
     * The tbl_vendor_route_product table is the one sent to the app during the login in order to have the relationship
     * between a product and what route the product has to use.
     * CAUTION !!! This function ASSUMES the tbl_vendor_list and tbl_vendor_route tables TO BE IN SYNC !
     * Be sure to execute this function not when the terminals log in.
     */
    app.post('/api/v1/rebuildVendorRouteProductTable', function (req, res) {
        var queryInsert = "";
        var locClinicId = "";

        if (res == null) {
            queryInsert = "insert into `tbl_vendor_route_product` (sap_material_id, vendor_material, vendor_id, vendor_route) \n " +
                "select sap_material_id, vendor_material, vendor_id, (select vendor_route from tbl_vendor_route where vendor_id = tbl_master_item.vendor_id) \n" +
                "from `tbl_master_item` where discontinued IS NULL or discontinued = ''";
            logger.info('API rebuildVendorRouteProductTable', 'Called with no clinicId');
        } else {
            locClinicId = res;
            queryInsert = "insert into `tbl_vendor_route_product` (sap_material_id, vendor_material, vendor_id, vendor_route) \n " +
                "select sap_material_id, vendor_material, vendor_id, \n" +
                "if (exists (select vendor_route from tbl_vendor_route_clinic where (vendor_id = tbl_master_item.vendor_id and clinic_id = '" + locClinicId + "')), \n" +
                "\t(select vendor_route from tbl_vendor_route_clinic where (vendor_id = tbl_master_item.vendor_id and clinic_id = '" + locClinicId + "')), \n" +
                "\t(select vendor_route from tbl_vendor_route where vendor_id = tbl_master_item.vendor_id))\n" +
                "from `tbl_master_item` where discontinued IS NULL or discontinued = ''";
            logger.info('API rebuildVendorRouteProductTable', 'Called with for the ' + locClinicId );
        }

        // First erase the current table
        logger.info('API rebuildVendorRouteProductTable', 'Erase table');
        var query = "delete from `tbl_vendor_route_product`";
        models.sequelize.query(query).spread(function (results, metadata) {

        }).catch(function (insErr) {
            logger.info('API rebuildVendorRouteProductTable', '- ERROR on delete - ' + insErr);
            res.sendStatus(500);
            return;
        });

        logger.info('API rebuildVendorRouteProductTable', 'Reset index');
        query = "alter table `tbl_vendor_route_product` auto_increment=1";
        models.sequelize.query(query).spread(function (results, metadata) {
        }).catch(function (insErr) {
            logger.info('API rebuildVendorRouteProductTable', '- ERROR on reset index - ' + insErr);
            res.sendStatus(500);
            return;
        });

        logger.info('API rebuildVendorRouteProductTable', 'Rebuild table');
        // Rebuild the table
        models.sequelize.query(queryInsert).spread(function (results, metadata) {
            res.sendStatus(200);
            return;
        }).catch(function (insErr) {
            logger.info('API rebuildVendorRouteProductTable', '- ERROR on insert - ' + insErr);
            res.sendStatus(500);
            return;
        });
    });

    /**
     * Wrapper API for the rebuild function
     */
    app.post('/api/v1/rebuildVendorRouteAppTable', function (req, res) {
        rebuildRoutingTable();
        res.sendStatus(200);
    });

    /**
     * Rebuild the tbl_vendor_route_app table - the API is usually called from webconsole when the tbl_vendor_route
     * or the tbl_vendor_route_clinic is changed
     * The tbl_vendor_route_app table is the one sent to the app during the login
     * CAUTION !!! This function ASSUMES the tbl_vendor_list and tbl_vendor_route tables TO BE IN SYNC !
     * CAUTION !!! Architecture changed because FRES-169 - be sure to run this function when NO LOGIN are running
     * The rebuild forcing flag is reset here if the routing table is successfully rebuilt in order to reduce the
     * number of rebuilding. i.e. if the rebuild is forced manually the flag is reset after the build is done.
     */
    function rebuildRoutingTable() {
        var query = "";
        var queryInsert = "";
        var queryDefaultRoute = "";

        queryInsert = "insert into `tbl_vendor_route_app` ( clinic_id, vendor_id, vendor_route)\n" +
            "select\n" +
            "\tc.clinic_id,\n" +
            "    a.vendor_id,\n" +
            "    if (exists (select distinct vendor_route from tbl_vendor_route_clinic where vendor_id = a.vendor_id and c.clinic_id = clinic_id),\n" +
            "    (select distinct vendor_route from tbl_vendor_route_clinic where vendor_id = a.vendor_id and c.clinic_id = clinic_id),\n" +
            "\ta.vendor_route )\n" +
            "\tfrom tbl_vendor_route as a, tmp_clinic_list as c;";

        queryDefaultRoute = "insert into `tbl_vendor_route_app` ( clinic_id, vendor_id, vendor_route) \n " +
            "select '.0000BM', vendor_id, vendor_route\n" +
            "from tbl_vendor_route";

        // Keep the sequence of operations nested !
        // First erase the current table
        logger.info('API rebuildVendorRouteAppTable', 'Erase table');
        query = "delete from `tbl_vendor_route_app`";
        models.sequelize.query(query).spread(function (results, metadata) {
            logger.info('API rebuildVendorRouteAppTable', 'Reset index');
            query = "alter table `tbl_vendor_route_app` auto_increment=1";
            models.sequelize.query(query).spread(function (results, metadata) {
                logger.info('API rebuildVendorRouteAppTable', 'Erase temporary table content');
                query = "delete from `tmp_clinic_list`";
                models.sequelize.query(query).spread(function (results, metadata) {
                    logger.info('API rebuildVendorRouteAppTable', 'Filling up temporary table');
                    query = "insert into tmp_clinic_list (clinic_id) select distinct clinic_id from tbl_vendor_route_clinic";
                    models.sequelize.query(query).spread(function (results, metadata) {
                        logger.info('API rebuildVendorRouteAppTable', 'Adding default record - fake clinic .0000BM');
                        // Rebuild the table - start with the default record
                        models.sequelize.query(queryDefaultRoute).spread(function (results, metadata) {
                            logger.info('API rebuildVendorRouteAppTable', 'Rebuild routing table');
                            models.sequelize.query(queryInsert).spread(function (results, metadata) {
                                logger.info('API rebuildVendorRouteAppTable', 'Rebuild completed - OK');

                                // FRES-128 - add entry in the activity log
                                writeEntryLog('tbl_vendor_route_app', 'Successfull rebuilt routing table');
                                global.buildRouting = false;    // Reset rebuilding flag
                                return true;
                            }).catch(function (insErr) {
                                // FRES-128 - add entry in the activity log
                                logger.info('API rebuildVendorRouteAppTable', '- ERROR on populating routing table - ' + insErr);
                                writeEntryLog('tbl_vendor_route_app', 'Error populating routing table');
                            });
                        }).catch(function (insErr) {
                            logger.info('API rebuildVendorRouteAppTable', '- ERROR on insert default - ' + insErr);
                        });
                    }).catch(function (insErr) {
                        logger.info('API rebuildVendorRouteAppTable', '- ERROR on populating temporary table - ' + insErr);
                    });
                }).catch(function (insErr) {
                    logger.info('API rebuildVendorRouteAppTable', '- ERROR on erasing temporary table - ' + insErr);
                });
            }).catch(function (insErr) {
                logger.info('API rebuildVendorRouteAppTable', '- ERROR on reset index - ' + insErr);
            });
        }).catch(function (insErr) {
            logger.info('API rebuildVendorRouteAppTable', '- ERROR on delete routing table - ' + insErr);
        });
        return false;
    }

    // -------------------------------------------------------------------  Tables management ---- END

    //...this is used by devices - currently seems unused - left for backward compatibility
    app.get('/api/getMasterItemList', function(req, res) {
        commonMasterItemList(req, res, true);
    });

    //...this is used by devices
    app.get('/api/v2/getMasterItemList', function(req, res) {
        commonMasterItemList(req, res, false);
    });

    function commonMasterItemList(req, res, isBackwards) {
        var date = req.query.date;
        var where = [];
        //logger.debug("getting master items at: " + new Date());

        if (date != null) {
            //logger.debug("getting master items for date: " + JSON.stringify(date));
            where.push({date_modified: {$gt: date}});
        }
        models.MasterItem.findAll({
                include: [{ model: models.CatalogList, as: 'Catalogs' }],
                where: where,
                paranoid: false
            }
        ).then(function (countRecords) {
            updateItemCatalog(countRecords, isBackwards);
            //logger.debug("sending master items: " + new Date());
            logger.log('trace', JSON.stringify(countRecords)) ;
            res.send(countRecords);
        });
    }

    //...this is used by the UI
    app.post('/api/v2/getMasterItem', function(req, res) {
        // TFG NOTE ! CAUTION !
        // Even if the freaking lint says to fix == into === or != into !== DON'T DO IT !
        if (req.body.sapMaterial == null) {
            if (req.body.where) {
                var args = buildMasterItemWhereArgs(req);
                models.MasterItem.findAll({ where: args, include: [{model: models.CatalogList, as: 'Catalogs'}]})
                    .then(function (items) {
                        updateItemCatalog(items);
                        res.send(items);

                    });
            } else {
                models.MasterItem.findAll({include: [{model: models.CatalogList, as: 'Catalogs'}]})
                    .then(function (items) {
                        updateItemCatalog(items);
                        res.send(items);

                    });
            }
        } else {
            models.MasterItem.findAll({
                include: [{ model: models.CatalogList, as: 'Catalogs' }],
                where: {
                    $and: [
                        {sapMaterial: req.body.sapMaterial},
                        {vendorMaterial: req.body.vendorMaterial}
                    ]
                }
            }).then(function (items) {
                updateItemCatalog(items, false);
                res.send(items);
            });
        }
    });

    function updateItemCatalog(items, isBackwards) {
        for (var i=0; i < items.length; i++) {
            var item = items[i];
            item.medVendor = "";
            var catalogs = item.Catalogs;

            if (isBackwards) {
                if (catalogs && catalogs.length > 0) {
                    item.medVendor = catalogs[0].catalogId;
                }
            } else {
                item.medVendor = item.catalogAsString;
            }
            //delete item.Catalogs
        }
        return items;
    }

    //CLINICS
    //this is used by devices.... but deprecated
    app.get('/api/getMobileClinicProfile', function(req, res) {
        var clinicId = req.query.clinic;
        var date = req.query.date;

        var where = [];
        if (date != null) {
            //logger.debug("getting clinic profile for date modified: " + JSON.stringify(date));
            where.push({date_modified: {$gt: date}});
        }
        if (clinicId != null) {
            // where.push(  );

            models.ClinicProfile.findAll({
                where: models.ClinicProfile.sequelize.where(models.ClinicProfile.sequelize.fn('dbo.fnRemoveNonNumericCharacters', models.ClinicProfile.sequelize.col('clinic_id')), clinicId)
                ,raw : true
            }).then(function (profile) {
                //old Android code is expecting one row with a numeric clinic id
                //var json = JSON.parse(JSON.stringify(profile));
                if (profile[0] != null) {
                    profile[0].clinicId = new Number(profile[0].clinicId.replace(/[^0-9]+/g, ''));
                    res.send(profile[0]);
                } else {
                    res.sendStatus(500);
                }
            });
        } else {
            res.sendStatus(500);
        }
    });

    //this is used by devices... and must be backwards compatible
    app.get('/api/v2/getMobileClinicProfile', function(req, res) {
        var clinicId = req.query.clinic;
        var date = req.query.date;

        var where = [];
        if (date != null) {
            //logger.debug("getting clinic profile for date modified: " + JSON.stringify(date));
            where.push({date_modified: {$gt: date}});
        }
        if (clinicId != null) {
            where.push({clinic_id: clinicId});
            models.ClinicProfile.findOne({
                include: [{ model: models.CatalogList, as: 'Catalogs' }],
                where: where
            }).then(function (profile) {
                updateClinicVendorList(profile);
                res.send(profile)
            });
        } else {
            res.sendStatus(500);
        }
    });


    function updateClinicCatalog(clinics) {
        if (clinics) {
            for (var i = 0; i < clinics.length; i++) {
                var clinic = clinics[i];
                updateClinicVendorList(clinic);
            }
        }
        return clinics;
    }

    function updateClinicVendorList(clinic){
        if (clinic) {
            clinic.vendorList = "";
            clinic.catalogList = "";
            var catalogs = clinic.Catalogs;
            var v = [];
            for (var c = 0; c < catalogs.length; c++) {
                v.push(catalogs[c].catalogId);
            }
            clinic.catalogList = v.join();
            clinic.vendorList = clinic.catalogList; //TODO: Make an actual Vendor list
        }
    }

    function buildDeviceCountQuery(location) {
        var table_name = "tbl_master_item";
        var vendor_fld = "catalog_id";
        var desc_fld="material_description";
        var cntType_fld = "count_type";
        var vl = location.vendorList;
        var vlList = [];
        if (vl && vl.length > 0) {
            if (vl.indexOf(",") > -1) {
                vlList = vl.split(/,/);
            } else {
                vlList.push(vl);
            }
        }
        var where = [];
        //first vendors
        if (vlList.length == 0) {
            where.push(table_name + "."+ vendor_fld + " = ''");
        }
        for (var v = 0; v < vlList.length; v++ ) {
            var vendor = vlList[v];
            if (!vendor || vendor.length == 0) {
                where.push("vl."+ vendor_fld + " = ''")
            } else {
                var escapedName = vendor; //models.sequelize.constructor.Utils.escape(vendor);
                where.push("vl." + vendor_fld + " = '" + escapedName + "'");
            }
        };
        var query = 'SELECT * FROM ' + table_name + " mi" +
            " WHERE mi.discontinued is null AND mi.cnt_ind = 1 AND mi." + cntType_fld + " IN ('ESA','Other') AND EXISTS (SELECT 1 FROM tbl_item_catalog ic, tbl_catalogs vl WHERE ic.item_id = mi.id AND vl.id = ic.catalog_id AND (";
        for (var v=0; v < where.length; v++) {
            var w = where[v];
            query += ((v==0) ? "" : " OR ") + w;
        };
        query += "))";
        where=[];
        //now item names
        if (!(location.mirceraVisible===true)) {
            where.push("mi." + desc_fld + " NOT LIKE 'MIRCERA%' ");
        }
        if (!(location.epogenVisible===true)) {
            where.push("mi." + desc_fld + " NOT LIKE 'EPOGEN%' ");
        }
        if (!(location.retacritVisible===true)) {
            where.push("mi." + desc_fld + " NOT LIKE 'RETACRIT%' ");
        }
        if (!(location.areanespVisible===true)) {
            where.push("mi." + desc_fld + " NOT LIKE 'ARANESP%' ");
        }
        if (where.length > 0) {
            query += " AND ( " + where[0];
            if (where.length > 1) {
                for (var w = 1; w < where.length; w++) {
                    query += " AND " + where[w];
                }
            }
            query += ")";
        }
        query += " ORDER BY mi." + cntType_fld + ", mi." + desc_fld;
        console.log(query);
        return query;
    }

    app.post('/api/getMasterItemAsDeviceCount', function(req, res) {
        var location = req.body.location;
        if (location) {
            var query = buildDeviceCountQuery(location);
            models.sequelize.query(query, { model: models.MasterItem}).then(function (items) {
                res.send(items);
            });
        }

    });

    app.post('/api/v2/getMasterItemAsDeviceCount', function(req, res) {
        var location = req.body.location;
        if (location) {
            var query = buildDeviceCountQuery(location);
            models.sequelize.query(query, { model: models.MasterItem}).then(function (items) {
                var io = new IOService(configJson, winston);
                var promises = [];
                var clinic = location.clinicId;
                for (var i=0; i< items.length; i++) {
                    var item = items[i];
                    var vendorMaterial = item.dataValues.vendor_material;
                    promises.push(io.getParValuePromise(clinic, vendorMaterial));
                    item.dataValues["min_val"] = 0;
                    item.dataValues["max_val"] = 0;
                }
                Promise.all(promises)
                    .catch(function(err){

                    }).then(function(results){
                    for(var r = 0; r< results.length; r++) {
                        var par = io.processResponse(results[r]);
                        if (par) {
                            var item = null;
                            if (par.vendorMaterial) {
                                for (var i = 0; i < items.length; i++) {
                                    if (items[i].dataValues.vendor_material == par.vendorMaterial) {
                                        item = items[i];
                                        break;
                                    }
                                }
                                if (item) {
                                    item.dataValues["min_val"] = par.minVal;
                                    item.dataValues["max_val"] = par.maxVal;
                                }
                            }
                        }
                    }
                    res.send(items);
                });
            });
        }

    });



    //Special API for add or update functionality
    app.post('/api/v2/saveMasterItem', function(req, res) {
        console.log("request: " + req.session);
        upsertMasterItem(req.body, req)
            .then(function(result) {
                if (!result) result = req.body;
                res.send(result);
            });
    });

    app.post('/api/deleteMasterItem', function(req, res) {
        models.MasterItem.destroy({
            where: { id: req.body.id }
        })
            .then(function (){
                logger.info("Master item deleted by: " + req.user.username + " record: " + JSON.stringify(req.body));
                res.send(req.body);
            });
    });

    app.post('/api/getClinicProfile', function(req, res) {
        if (req.body.clinicId == null) {
            models.ClinicProfile.findAll().then(function(clinicProfiles) {
                res.send(clinicProfiles);
            });
        } else {
            models.ClinicProfile.findAll({
                where: {
                    clinicId: req.body.clinicId
                }

            }).then(function (clinicProfiles) {
                res.send(clinicProfiles);
            });
        }
    });

    app.post('/api/v2/getClinicProfile', function(req, res) {
        if (req.body.clinicId == null) {
            if (req.body.where) {
                var args = buildClinicProfileWhereArgs(req);
                var catalogArg = null;
                var delIndex = -1;
                for(var i = 0; i < args.length; i++) {
                    var it = args[i];
                    if (_.findIndex(Object.keys(it), function(o) { return o == "Catalogs" ; }) >=0) {
                        delIndex = i;
                        catalogArg = {};
                        catalogArg['catalog_id'] = it['Catalogs'];
                        break;
                    }
                }
                if (delIndex >= 0)
                {
                    args.splice(delIndex,1);
                }
                models.ClinicProfile.findAll({where: args, include: [{model: models.CatalogList, as: 'Catalogs'}]})
                    .then(function (clinicProfiles) {
                        var toSend = updateClinicCatalog(clinicProfiles);
                        if (catalogArg) {
                            var realArg = String(catalogArg['catalog_id']['$like']).replace(/\%/g,"");
                            toSend = _.filter(clinicProfiles, function(it) {    return (it.catalogList.indexOf(realArg) >= 0); });
                        }
                        res.send(toSend);
                    });

            } else {
                models.ClinicProfile.findAll({
                    include: [{
                        model: models.CatalogList,
                        as: 'Catalogs'
                    }]
                }).then(function (clinicProfiles) {
                    updateClinicCatalog(clinicProfiles);
                    res.send(clinicProfiles);
                });
            }
        } else {
            models.ClinicProfile.findAll({
                include: [{ model: models.CatalogList, as: 'Catalogs' }],
                where: {
                    clinicId: req.body.clinicId
                }

            }).then(function (clinicProfiles) {
                updateClinicCatalog(clinicProfiles);
                res.send(clinicProfiles);
            });
        }
    });

    app.post('/api/deleteClinicProfile', function(req, res) {
        var clinicId = req.body.clinicId;
        if (clinicId) {
            models.ClinicProfile.destroy(
                {where: {clinicId: clinicId}}
            ).then(function () {
                logger.info("Clinic profile deleted by: " + ((req.user && req.user.username) ? req.user.username : "tester") + " record: " + JSON.stringify(req.body));
                res.send(req.body);
            });
        } else {
            res.status(200).send({});
        }
    });

    app.post('/api/v2/saveClinicProfile', function(req, res) {
        upsertClinicProfile(req.body, req)
            .then(function(result) {
                if (!result) result = req.body;
                res.send(result);
            });
    });

    /**
     * Adds a specified vendorId to every clinic's list
     *   that does not already have it.
     *
     * NOTE: The Vendor ID MUST exist in the Vendor table
     *
     * Sample request bodies: { "vendorId": "value" }   OR   { "vendorId": "-BLANK-" }
     **/
    app.post('/api/bulkAddClinicVendor', function(req, res) {
        console.log("bulkAddClinicVendor  BODY:" + req.body.vendorId);
        models.VendorList.findAll({
            where: {
                $or: [
                    { vendorId : req.body.vendorId},
                    { vendorDisplay: req.body.vendorId }
                ]
            },raw:true})
            .then(function(vendors) {
                if (vendors && vendors.length > 0) {
                    var table_name = "[tbl_clinic_profile]";
                    var vendor_fld = "[vendor_list]";
                    var query = "UPDATE " + table_name + " SET " + vendor_fld + " = " + vendor_fld + " + '," + vendors[0].vendorId + "'";
                    query += " WHERE " + vendor_fld + " NOT LIKE '" + vendors[0].vendorId + ",%'"; //starts with
                    query += " AND " + vendor_fld + " NOT LIKE '%," + vendors[0].vendorId + "'"; //ends with
                    query += " AND " + vendor_fld + " NOT LIKE '%," + vendors[0].vendorId + ",%'"; //contains middle
                    query += " AND " + vendor_fld + " <> '" + vendors[0].vendorId + "'"; //only entry
                    console.log(query);
                    models.sequelize.query(query).spread(function(results, metadata) {
                        res.send(results);
                    });
                } else {
                    res.sendStatus(500);
                }
            }, function(err) {
                logger.error(err);
                res.status = 500;
                res.send(err);
            });
    });

    app.post('/api/v2/bulkAddClinicVendor', function(req, res) {
        console.log("bulkAddClinicVendor  BODY:" + req.body.vendorId);
        models.CatalogList.findAll({
            where: {
                $or: [
                    { catalogId : req.body.catalogId},
                    { catalogDisplay: req.body.catalogId }
                ]
            },raw:true})
            .then(function(vendors) {
                if (vendors && vendors.length > 0) {
                    var table_name = "[tbl_clinic_catalog]";
                    var clinic_tbl = "[tbl_clinic_profile]";
                    for (var v=0; v < vendors.length; v++) {
                        var vendor = vendors[v];
                        var reset = "DELETE FROM " + table_name + " WHERE catalog_id = " + vendor.id;
                        var query = "INSERT INTO " + table_name + " (clinic_id, catalog_id) SELECT ID, " + vendor.id + " FROM  " + clinic_tbl;
                        console.log(query);
                        models.sequelize.query(reset).spread(function(resets, metadata) {
                            models.sequelize.query(query).spread(function(results, metadata) {
                                res.send(results);
                            }).catch(function(insErr) {
                                logger.error(insErr);
                                res.status = 500;
                                res.send(insErr);
                            });
                        }).catch(function(delErr) {
                            logger.error(delErr);
                            res.status = 500;
                            res.send(delErr);
                        });
                    }
                } else {
                    res.sendStatus(500);
                }
            }).catch(function(err) {
            logger.error(err);
            res.status = 500;
            res.send(err);
        });
    });

    /**
     * Removes a specified vendorId in every clinic's list
     *   that already has it
     *
     * NOTE: The Vendor ID is NOT checked against the Vendor table
     *
     * Sample request bodies: { "vendorId": "value" }   OR   { "vendorId": "-BLANK-" }
     **/
    app.post('/api/bulkRemoveClinicVendor', function(req, res) {
        if (req.body && req.body.vendorId) {
            if (req.body.vendorId) {
                var vendor = req.body.vendorId;
                if (vendor == "-BLANK-") { vendor = ""; }
                var table_name = "[tbl_clinic_profile]";
                var vendor_fld = "[vendor_list]";
                var query = "UPDATE " + table_name + " SET " + vendor_fld + " = (CASE  ";
                query += " WHEN " + vendor_fld + " LIKE '" + vendor + ",%'   THEN REPLACE(" + vendor_fld + ", '" + vendor + ",','') "; //starts with
                query += " WHEN " + vendor_fld + " LIKE '%," + vendor + "'   THEN REPLACE(" + vendor_fld + ", '," + vendor + "','') "; //ends with
                query += " WHEN " + vendor_fld + " LIKE '%," + vendor + ",%' THEN REPLACE(" + vendor_fld + ", '," + vendor + "','') "; //contains middle
                query += " WHEN " + vendor_fld + " = '" + vendor + "' THEN '' "; //only entry
                query += " ELSE " + vendor_fld + " END) ";
                query += " WHERE " + vendor_fld + " LIKE '" + vendor + ",%'"; //starts with
                query += " OR " + vendor_fld + "  LIKE '%," + vendor + "'"; //ends with
                query += " OR " + vendor_fld + "  LIKE '%," + vendor + ",%'"; //contains middle
                query += " OR " + vendor_fld + " = '" + vendor + "'"; //only entry

                models.sequelize.query(query).spread(function(results, metadata) {
                    res.send(results);
                }, function(err) {
                    logger.error(err);
                    res.status = 500;
                    res.send(err);
                });
            } else {
                res.sendStatus(500);
            }
        };
    });

    /**
     * Removes a specified vendorId in every clinic's list
     *   that already has it
     *
     * NOTE: The Vendor ID MUST BE checked against the Vendor table
     *
     * Sample request bodies: { "vendorId": "value" }   OR   { "vendorId": "-BLANK-" }
     **/
    app.post('/api/v2/bulkRemoveClinicVendor', function(req, res) {
        if (req.body && req.body.vendorId) {
            if (req.body.vendorId) {
                var vendor = req.body.vendorId;
                var table_name = "[tbl_clinic_catalog]";
                var vendor_table_name = "[tbl_catalogs]";
                var vendor_fld = "[catalog_id]";
                var query = "DELETE FROM " + table_name;
                query += " WHERE " + vendor_fld + " IN (SELECT ID FROM " + vendor_table_name + " WHERE " + vendor_fld + " = '" + vendor + "'";
                query += " OR [catalog_display] = '" + vendor + "')";

                models.sequelize.query(query).spread(function(results, metadata) {
                    res.send(results);
                }).catch(function(err) {
                    logger.error(err);
                    res.status = 500;
                    res.send(err);
                });
            } else {
                res.sendStatus(500);
            }
        }
    });

    /**
     *  Update or Insert a record. Searches for a record, if found, updates it. If not found, creates it then updates it.
     *  Returns 1 if updated, 0 if created, -1 for error (but not really)
     */
    function upsertClinicProfile(clinicProfile, req) {
        if (clinicProfile.hasOwnProperty("Catalogs")) delete clinicProfile.Catalogs;
        return models.ClinicProfile.findOrCreate({where: {clinicId: clinicProfile.clinicId}, raw: true})
            .spread(function(clinicRecord, created) {
                clinicProfile.id = clinicRecord.id;
                models.ClinicProfile.update(clinicProfile, { where: { id: clinicProfile.id }})
                    .then(function(count) {
                        if (count > 0 && clinicProfile.vendorList) {
                            var ids = clinicProfile.vendorList.split(",");
                            models.ClinicCatalog.destroy({ where: { clinicId: clinicProfile.id }}).then(function(num){
                                var cats = [];
                                for (var i = 0; i < ids.length;i++) {
                                    var cat = models.ClinicCatalog.build({
                                        clinicId: clinicProfile.id,
                                        catalogId: ids[i]
                                    });
                                    cats.push(cat);
                                }
                                if (cats.length > 0) {
                                    var promises = [];

                                    cats.forEach(function(cat) {
                                        promises.push(cat.save());
                                    });

                                    models.sequelize.Promise.all(promises).then(function(vals){
                                        logger.info(process.pid + ":Clinic profile " + ((created) ? "created" : "updated") + " by: " + ((req.user && req.user.username) ? req.user.username : "tester") + " record: " + JSON.stringify(clinicProfile));
                                        return cats.length;
                                    }).catch(function(saveErr){
                                        logger.error(saveErr);
                                        return -1;
                                    });
                                }
                            })
                        } else {
                            logger.info("Clinic profile " + ((created) ? "created" : "updated") + " by: " + ((req.user && req.user.username) ? req.user.username : "tester") + " record: " + JSON.stringify(clinicProfile));
                            return count;
                        }
                    }).catch(function(err) {
                    logger.error(err);
                    return -1;
                });
            }).catch(function(err) {
                logger.error(err);
                return -1;
            });
    }

    /**
     *  Update or Insert a record. Searches for a record, if found, updates it. If not found, creates it then updates it.
     *  Returns 1 if updated, 0 if created, -1 for error (but not really)
     */
    function upsertMasterItem(masterItem, req) {
        return models.MasterItem.findOrCreate({where: {sapMaterial: masterItem.sapMaterial, vendorMaterial: masterItem.vendorMaterial}, raw: false, paranoid: false })
            .spread(function(itemRecord, created) {
                masterItem.id = itemRecord.id;
                if (!created && itemRecord.discontinued) {
                    itemRecord.deletedAt = null;
                    itemRecord.restore();
                }
                if (masterItem.hasOwnProperty("Catalogs")) delete masterItem.Catalogs;
                models.MasterItem.update(masterItem, {where: {id: masterItem.id}})
                    .then(function (count) {
                        if (count > 0 && masterItem.medVendor) {
                            var ids = masterItem.medVendor.split(",");
                            models.ItemCatalog.destroy({where: {itemId: masterItem.id}}).then(function (num) {
                                var cats = [];
                                for (var i = 0; i < ids.length; i++) {
                                    var cat = models.ItemCatalog.build({
                                        itemId: masterItem.id,
                                        catalogId: ids[i]
                                    });
                                    cats.push(cat);
                                }
                                if (cats.length > 0) {
                                    var promises = [];

                                    cats.forEach(function (cat) {
                                        promises.push(cat.save());
                                    });

                                    models.sequelize.Promise.all(promises).then(function (vals) {
                                        logger.info(process.pid + ":Master Item " + ((created) ? "created" : "updated") + " by: " + ((req.user && req.user.username) ? req.user.username : "tester") + " record: " + JSON.stringify(masterItem));
                                        return cats.length;
                                    }).catch(function (saveErr) {
                                        logger.error(saveErr);
                                        return -1;
                                    });
                                }
                            }).catch(function (err) {
                                logger.error(err);
                                return -1;
                            });
                        }
                    }).catch(function (err) {
                    logger.error(err);
                    return -1;
                });
            }).catch(function(err) {
                logger.error(err);
                return -1;
            });
    }

    /**
     *  Update or Insert each inventory or count record. Searches for a record, if found, updates it. If not found, creates it then updates it.
     *  Returns all promises (to ensure caller waits for it to finish)
     */
    function bulkCreateData(dataObjects, type) {

        var promises = [];

        if (type == 'inventory') {
            if (dataObjects != null) {
                dataObjects.forEach(function(record) {
                    logger.debug(process.pid + ":saving inventory record: " + JSON.stringify(record));
                    record.id = null;
                    if (record.asnId && record.asnId == 0) { record.asnId = null; }
                    promises.push(models.Inventory.findOrCreate({where: {deviceId: record.deviceId, timestamp: record.timestamp, type: record.type, vendorMaterial: record.vendorMaterial }, defaults: {deviceId: record.deviceId, timestamp: record.timestamp,type: record.type, clinicId: record.clinicId, count: record.count, vendorMaterial: record.vendorMaterial, asnId: record.asnId }, raw: true})
                        .spread(function(inv, created) {
                            record.id = inv.id;
                            models.Inventory.update(record, { where: { id: record.id }});
                            // THE FOLLOWING WAS REPLACED WITH A TRIGGER.  CODE KEPT IN CASE TRIGGERS REJECTED BY FRESENIUS.
                            // if (record.asnId && record.asnId > 0) {
                            //     var asnRequest = {
                            //         body: {
                            //             id: record.asnId,
                            //             receivedQty: record.count
                            //         }
                            //     }
                            //     var helper = new AsnHelper(configJson);
                            //     helper.getAsnUpdatePromise(asnRequest).then(function(asnResponse){
                            //         logger.info("asn updated")
                            //     }).catch(function(aerr) {
                            //         logger.error(JSON.stringify(aerr));
                            //     });
                            // }
                            return {
                                id: record.id,
                                devicePk: record.devicePk
                            };
                        }).catch(function(err) {
                            logger.error(err);
                        })
                    );
                });
            }
        } else if (type == 'count') {
            if (dataObjects != null) {
                dataObjects.forEach(function (countData) {
                    logger.debug(process.pid + ":saving count record: " + JSON.stringify(countData));
                    countData.id = null;
                    promises.push(models.CountData.findOrCreate({
                            where: {
                                deviceId: countData.deviceId,
                                timestamp: countData.timestamp,
                                vendorMaterial: countData.vendorMaterial
                            }, raw: true
                        }).spread(function (cd, created) {
                            countData.id = cd.id;
                            models.CountData.update(countData, {where: {id: countData.id}});
                            return {
                                id: countData.id,
                                devicePk: countData.devicePk
                            };
                        },
                        function (err) {
                            logger.error(err);
                        }
                        )
                    );
                });
            }
        }
        return models.sequelize.Promise.all(promises);
    }

    /**
     * Processes FTP log JSON as a list
     *
     * @param dataObjects
     * @returns {*}
     */
    function bulkCreateFTPData(dataObjects) {

        var promises = [];

        if (dataObjects != null) {
            dataObjects.forEach(function (ftpData) {
                //logger.debug("saving ftp record: " + JSON.stringify(ftpData));
                ftpData.id = null;
                promises.push(models.FTPLog.findOrCreate({
                        where: {
                            deviceId: ftpData.deviceId,
                            sent: ftpData.sent,
                            userId: ftpData.userId,
                            success: ftpData.success
                        }, raw: true
                    })
                        .spread(function (fd, created) {
                                ftpData.id = fd.id;
                                models.FTPLog.update(ftpData, {where: {id: ftpData.id}});
                                return {id : ftpData.id};
                            },
                            function (err) {
                                logger.error(err);
                            }
                        )
                );
            });
        }
        return models.sequelize.Promise.all(promises);
    }

    app.post('/api/saveFTPData', function(req, res) {
        bulkCreateFTPData(req.body)
            .done(function(data) {
                res.send(data);
            });
    });

    app.post('/api/saveVendor', function(req, res) {
        if (req.body.id && req.body.id > 0) {
            models.VendorList.update(req.body, {where: {
                    id: req.body.id
                }}).then(function(count){
                res.send(req.body);
            }).catch(function (err) {
                logger.error(err);
                res.sendStatus(500);
            });
        } else {
            if (req.body.vendorId) {
                models.VendorList.findOrCreate({
                    where: {
                        vendorId: req.body.vendorId
                    }, defaults: {
                        vendorDisplay: req.body.vendorDisplay,
                        minimumOrderAmount: req.body.minimumOrderAmount
                    }
                }).spread(function (data, created) {
                    if (!created) {
                        data.minimumOrderAmount = req.body.minimumOrderAmount;
                        data.vendorDisplay = req.body.vendorDisplay;
                        data.save().then(function () {
                            res.send(data);
                        }).catch(function (error) {
                            logger.error(error);
                            res.sendStatus(500);
                        });
                    } else {
                        res.send(data);
                    }
                }).catch(function (err) {
                    logger.error(err);
                    res.sendStatus(500);
                });
            } else {
                res.sendStatus(500);
            }
        }
    });

    app.post('/api/deleteVendor', function(req, res) {
        if (req.body.vendorId) {
            //TODO: make sure no Clinic or Item has this vendor assigned
            models.VendorList.destroy({
                where: req.body
            })
                .then(function (){
                    logger.info(process.pid + ":Vendor item deleted by: " + req.user.username + " record: " + JSON.stringify(req.body));
                    res.send(req.body);
                });
        }
        else {
            res.sendStatus(500);
        }
    });

    app.post('/api/saveCatalog', function(req, res) {
        if (req.body.id && req.body.id > 0) {
            models.CatalogList.update(req.body, {where: {
                    id: req.body.id
                }}).then(function(count){
                res.send(req.body);
            }).catch(function (err) {
                logger.error(err);
                res.sendStatus(500);
            });
        } else {
            if (req.body.catalogId) {
                models.CatalogList.findOrCreate({
                    where: {
                        catalogId: req.body.catalogId
                    }, defaults: {
                        catalogDisplay: req.body.catalogId,
                        isDefault: req.body.isDefault
                    }
                }).spread(function (data, created) {
                    if (!created) {
                        data.isDefault = req.body.isDefault;
                        data.vendorDisplay = req.body.catalogId;
                        data.save().then(function () {
                            res.send(data);
                        }).catch(function (error) {
                            logger.error(error);
                            res.sendStatus(500);
                        });
                    } else {
                        res.send(data);
                    }
                }).catch(function (err) {
                    logger.error(err);
                    res.sendStatus(500);
                });
            } else {
                res.sendStatus(500);
            }
        }
    });

    app.post('/api/deleteCatalog', function(req, res) {
        if (req.body.catalogId) {
            //TODO: make sure no Clinic or Item has this vendor assigned
            models.CatalogList.destroy({
                where: req.body
            })
                .then(function (){
                    logger.info(process.pid + ":Catalog item deleted by: " + req.user.username + " record: " + JSON.stringify(req.body));
                    res.send(req.body);
                });
        }
        else {
            res.sendStatus(500);
        }
    });


    /**   MESSAGING   **/
    app.get('/api/messages/:clinic', function(req, res) {
        if (req.params && req.params['clinic']){
            var inClinic = req.params['clinic'];
            models.ClinicProfile.findOne({
                where: { clinicId: inClinic }
            }).then(function(profile){
                if (profile && profile.id) {
                    var now = new Date();
                    models.MsgRecipient.findAll({
                        include: [{model: models.Message, as: 'message',
                            where: {
                                $or: [
                                    { keepAliveStart: { $eq: null } },
                                    {
                                        $and: {
                                            keepAliveStart: {$lte: now},
                                            keepAliveEnd: {$gte: now}
                                        }
                                    }]
                            }
                        }],
                        where: {
                            clinicId: profile.id
                        }
                    }).then(function (messages) {
                        var responseList = [];
                        if (messages && messages.length > 0) {
                            for (var r = 0; r < messages.length; r++) {
                                var item = messages[r];
                                if (item.timestamp || item.status === 'read') {
                                    if (!item.message.keepAliveStart) continue;
                                }
                                responseList.push(item.message);
                                item.status = "read";
                                item.timestamp = new moment();
                                item.save().catch(function(saveErr){
                                    console.log(saveErr);
                                });
                            }
                        }
                        res.send(responseList);
                    }).catch(function (msgErr) {
                        res.status(500).send("Error querying messages: " + msgErr);
                    });
                } else {
                    res.status(404).send("Clinic not found");
                }
            }).catch(function(clinicErr){
                res.status(404).send("Clinic not found");
            })
        } else {
            res.sendStatus(400);
        }
    });

    /**
     * Processes NDC scan JSON as a list
     *
     * @param dataObjects
     * @returns {*}
     */
    function bulkCreateNdcData(dataObjects) {

        var promises = [];

        if (dataObjects != null) {
            dataObjects.forEach(function (ndcData) {
                //logger.debug("saving ndc record: " + JSON.stringify(ndcData));
                ndcData.id = null;
                promises.push(models.NdcData.findOrCreate({
                        where: {
                            deviceId: ndcData.deviceId,
                            timestamp: ndcData.timestamp
                        }, raw: true
                    }).spread(function (fd, created) {
                        ndcData.id = fd.id;
                        models.NdcData.update(ndcData, {where: {id: ndcData.id}});
                        return {id : ndcData.id};
                    },
                    function (err) {
                        logger.error(err);
                    }
                    )
                );
            });
        }
        return models.sequelize.Promise.all(promises);
    }

    app.post('/api/saveNdcData', function(req, res) {
        bulkCreateNdcData(req.body)
            .done(function(data) {
                res.send(data);
            });
    });

    /**
     * Devices to get their ASNs
     */
    app.get('/api/v1/asn/:clinic', function(req,res){
        if (!req.params.clinic) {
            res.status(400).send("Missing Clinic ID");
        } else {
            var helper = new AsnHelper(configJson);
            helper.getOpenAsns(req.params.clinic).then(function (response) {
                res.header("X-Total-Count", response.headers['x-total-count']);
                res.status(200).send(response.body);
            }).catch(function (err) {
                res.status(500).send(err);
            });
        }
    });

    app.get('/api/v1/receipts/SAP', function(req, res){
        models.Inventory.findAll({
            include: [{model: models.InventoryASN, as: "ASN"}],
            where: {
                type: 'RECEIPT',
                sapFlag: 0,
                timestamp: {
                    $gte: new moment("2017-01-01", "YYYY-MM-DD")
                }
            }
        })
            .then(function(data){
                var out = [];
                if (data && Array.isArray(data) && data.length > 0) {

                    _.forEach(data, function(rec){
                        if (rec.ASN) {
                            var record = {
                                id: rec.id,
                                clinicId: rec.clinicId,
                                vendorMaterialId: rec.vendorMaterial,
                                sapMaterialId: rec.sapMaterial,
                                description: rec.description,
                                poNumber: rec.poPrefix + "-" + rec.poNumber,
                                receivedDate: rec.timestamp,
                                receivedQty: rec.count,
                                lot: (rec.lot) ? rec.lot : null,
                                expiration: (rec.expiration) ? rec.expiration : null,
                                asn: (rec.ASN) ? rec.ASN : null
                            };
                            out.push(record);
                        }
                    });
                }
                res.status(200).send(out);

            }).catch(function(err){
            res.status(500).send(err);
        });
    });

    app.put('/api/v1/receipts/SAP', function(req, res){
        if (req.body) {
            if (Array.isArray(req.body)) {
                var batches = _.chunk(req.body, 255);
                var total = 0;
                var promises = [];
                _.forEach(batches, function(batch) {
                    promises.push(
                        models.Inventory.update({sapFlag: 1}, {
                            where: {id: {$in: batch}}
                        })
                    );
                });
                models.sequelize.Promise.all(promises)
                    .then(function (cnts, rows) {
                        _.forEach(cnts, function(cnt) {
                            total += cnt;
                        });
                        res.status(200).json(total);
                    }).catch(function (errs) {
                    res.status(500).send(errs);
                });
            } else {
                models.Inventory.update({sapFlag: 1}, { where: {id: req.body }})
                    .spread(function(cnt, rows) {
                        res.status(200).json(cnt);
                    }).catch(function(err){
                    res.status(500).send(err);
                });
            }
        } else {
            res.status(400).send("No receipts to update");
        }
    });

    /**
     * Returns all Vendors in the table
     */
    app.get('/api/v1/getVendorsForDevice', function(req, res) {
        models.VendorList.findAll({order: '"vendorDisplay" ASC'}).then(function(vendors) {
            res.send(vendors);
        });
    });

    function buildWhereArgs(req, objName, objRawAttributes) {
        var whereArgs = [];
        if (req.body.where != null) {
            console.log(JSON.stringify(req.body.where));
            for (var key in req.body.where) {
                if (objRawAttributes.hasOwnProperty(key)) {
                    var newCondition = String(req.body.where[key]);
                    if (newCondition === "[object Object]") {
                        if (key === "dateModified") {
                            //we have gt/lt range
                            if (req.body.where[key]['$gt']) {
                                var newWhere = {};
                                newWhere[objRawAttributes[key].field] = {$gt: moment(req.body.where[key]['$gt'])};
                                whereArgs.push(newWhere);
                            }
                            if (req.body.where[key]['$lt']) {
                                var newWhere = {};
                                newWhere[objRawAttributes[key].field] = {$lt: moment(req.body.where[key]['$lt'])};
                                whereArgs.push(newWhere);
                            }
                            continue;
                        } else {
                            newCondition = String(req.body.where[key]['$like']);
                        }
                    }
                    var dataType = String(objRawAttributes[key].type);
                    if (newCondition.indexOf("*") < 0) {
                        newCondition = "%" + newCondition + "%";
                    } else {
                        newCondition = newCondition.replace(/\*/g, "%");
                    }
                    var whereArg = null;
                    if (dataType.indexOf("VARCHAR") < 0) {
                        //cast numerics as strings so we can use the "LIKE" operator
                        if (dataType.indexOf("TIME") > 0) {
                            //since time is hidden in the UI, we should strip time from the date fields
                            whereArg = models.Sequelize.where(models.Sequelize.fn('DATE_FORMAT', models.Sequelize.col(objName + "." + objRawAttributes[key].field), '%m-%d-%Y'), 'LIKE', newCondition);
                        } else if (dataType.indexOf("BOOLEAN") >= 0 || dataType.indexOf("TINYINT") >= 0) {
                            whereArg = {};
                            whereArg[objRawAttributes[key].field] = ((newCondition.charAt(1).toUpperCase() === 'T') ? 1 : 0);
                        } else if (dataType.indexOf("VIRTUAL") >= 0) {
                            //skip
                            whereArg = {};
                            whereArg['Catalogs'] = {$like: newCondition};
                        } else {
                            whereArg = {};
                            whereArg[objRawAttributes[key].field] = {$like: newCondition};
                        }
                    } else {
                        //already a String... just use the LIKE operator
                        whereArg = {};
                        whereArg[objRawAttributes[key].field] = {$like: newCondition};
                    }
                    whereArgs.push(whereArg);
                }
            }
        }
        return whereArgs;
    }

    function buildMasterItemWhereArgs(req) {
        return buildWhereArgs(req, 'MasterItem', models.MasterItem.rawAttributes);
    }
    function buildClinicProfileWhereArgs(req) {
        return buildWhereArgs(req, 'ClinicProfile', models.ClinicProfile.rawAttributes);
    }

    // ----- Activity log management

    /**
     * Wrapper API for the rebuild function
     */
    app.post('/api/purgeActivityLog', function (req, res) {
        purgeOldActivityLogEntries();
        res.sendStatus(200);
    });
    /**
     * Purge tbl_activity_log of entries older than 182 days (approx 6 months)
     * The function is called every day at 2 AM ET (see timer also used for rebuild routing table)
     * or via an API fomr the Activuty Log tab (if boss)
     */
    function purgeOldActivityLogEntries() {
        var queryRemove = "delete from tbl_activity_log where date_modification < (NOW() - INTERVAL 182 DAY) limit 20";

        models.sequelize.query(queryRemove).spread(function (results, metadata) {
            logger.info('purgeOldActivityLogEntries', 'Purged 20 Activity log table from records older than 182 days');
            writeEntryLog('tbl_activity_log', 'Purged 20 old activity log entries');
        }).catch(function (insErr) {
            logger.info('purgeOldActivityLogEntries', '- ERROR on purge Activity log table - ' + insErr);
            writeEntryLog('tbl_activity_log', 'FAILED Purging 20 old activity log entries');
        });
    }

    // Function to be used locally to write on the activity log
    function writeEntryLog(name_table, desc) {
        var record = {bodyId: 0, username: "system", nameTable: "", description: "", dateModified: 0};
        record.dateModified = new Date();
        record.nameTable = name_table;
        record.description = desc;

        logger.debug("writeEntryLog called - record: " + JSON.stringify(record, null, 4));

        try {
            if (record.nameTable && record.description) {
                models.ActivityLog.findOrCreate({
                    where: {
                        id: 0
                    }, defaults: {
                        userId: record.username,
                        nameTable: record.nameTable,
                        description: record.description,
                        modified: record.dateModified
                    }
                }).spread(function (data, created) {
                    logger.debug("   new record");     // TEMP

                    if (!created) {
                        data.userId = record.username;
                        data.nameTable = record.nameTable;
                        data.description = record.description;
                        data.modified = record.dateModified;

                        logger.debug("userId : " + data.userId);     // TEMP
                        logger.debug("nameTable : " + data.nameTable);     // TEMP
                        logger.debug("description : " + data.description);     // TEMP
                        logger.debug("modified : " + data.modified);     // TEMP

                        data.save().then(function () {
                            // Operation successfull
                        }).catch(function (error) {
                            // Operation failed
                            logger.error(error);
                        });
                    } else {
                        // Operation successful
                    }
                }).catch(function (err) {
                    // Problems
                    logger.error(err);
                });
            }
        } catch (e) {
            logger.error(e);
        }
    }
};
